from .GoogleTokenSpan import GoogleTokenSpan
from .GoogleSentiment import GoogleSentiment


class GoogleMention(GoogleTokenSpan):
	def __init__(self, dictionary, document, entity):
		text = dictionary.pop('text')
		content = text.pop('content')
		begin = text.pop('begin_offset')
		end = begin + len(content)
		super().__init__(dictionary=dictionary, document=document, begin=begin, end=end)
		self._text = content
		self._index = None
		self._entity = entity

	@property
	def entity(self):
		"""
		:rtype: GoogleEntity
		"""
		return self._entity

	@property
	def id(self):
		return self.document.id, 'mention', self.entity._index, self._index

class GoogleEntity:
	def __init__(self, dictionary, document):
		self._dictionary = dictionary
		self._document = document
		self._name = self._dictionary.pop('name')
		self._type = self._dictionary.pop('type')
		self._metadata = self._dictionary.pop('metadata')
		self._wikipedia_url = self._metadata.pop('wikipedia_url', None)
		self._salience = self._dictionary.pop('salience')
		sentiment = self._dictionary.pop('sentiment')
		self._sentiment = GoogleSentiment(score=sentiment.pop('score'), magnitude=sentiment.pop('magnitude'))
		self._mentions = [
			GoogleMention(dictionary=mention, document=self.document, entity=self)
			for mention in self._dictionary.pop('mentions')
		]
		for index, mention in enumerate(self.mentions):
			mention._index = index
		self._index = None

	def graph_str(self):
		return f'{self.name}\n({str(self._type).replace("_", " ")})'

	@property
	def id(self):
		return self.document.id, 'entity', self._index

	def __str__(self):
		return f'{self.name} ({self._type})'

	def __repr__(self):
		return str(self)

	@property
	def document(self):
		"""
		:type: .GoogleDocument.GoogleDocument
		"""
		return self._document

	@property
	def mentions(self):
		"""
		:rtype: list[GoogleMention]
		"""
		return self._mentions

	@property
	def dictionary(self):
		"""
		:rtype: dict
		"""
		return self._dictionary

	@property
	def name(self):
		return self._name

	@property
	def type(self):
		return self._type

	@property
	def salience(self):
		return self._salience

	@property
	def sentiment(self):
		"""
		:rtype: GoogleSentiment
		"""
		return self._sentiment

	@property
	def metadata(self):
		"""
		:rtype: dict or NoneType
		"""
		return self._metadata

	@property
	def wikipedia_url(self):
		"""
		:rtype: str or NoneType
		"""
		return self._wikipedia_url

	@property
	def tokens(self):
		return [token for mention in self.mentions for token in mention.tokens]



from .analyze import analyze
from .GoogleDocument import GoogleDocument


class GoogleCloud:
	def __init__(
			self,
			api_key,
			language='english',
			standardize=True,
			extract_entities=True,
			extract_sentiment=True,
			extract_syntax=True,
			cache=None
	):
		"""
		:param str api_key:
		:param str language:
		:param bool standardize:
		:param bool extract_entities:
		:param bool extract_sentiment:
		:param bool extract_syntax:
		:param disk.Cache.Cache cache:
		"""
		self._api_key = api_key
		self._language = language[:2].lower()
		self._standardize = standardize
		self._extract_entities = extract_entities
		self._extract_sentiment = extract_sentiment
		self._extract_syntax = extract_syntax
		self._cache = cache
		if self._cache:
			self._cached_analyze = self._cache.make_cached(
				function=self._analyze, id='google_cloud_analyze_function',
				sub_directory='analyze'
			)
		else:
			self._cached_analyze = self._analyze

	def _analyze(
			self, text, language, standardize, extract_entities,
			extract_sentiment, extract_syntax
	):
		return analyze(
			text=text, api_key=self._api_key, language=language, standardize=standardize,
			extract_entities=extract_entities, extract_sentiment=extract_sentiment,
			extract_syntax=extract_syntax
		)

	def analyze(self, text):
		return self._cached_analyze(
			text=text,
			language=self._language,
			standardize=self._standardize,
			extract_entities=self._extract_entities,
			extract_sentiment=self._extract_sentiment,
			extract_syntax=self._extract_syntax
		)

	def create_document(self, text):
		analysis = self.analyze(text=text)
		return GoogleDocument(cloud=self, _analysis=analysis)




from .BertVectorizerModel import BertVectorizerModel
import torch


class BertVectorizer:
	def __init__(self, num_tokens=50):
		self._has_gpu = torch.cuda.device_count() > 0
		self._model = BertVectorizerModel(num_tokens=num_tokens)

		if self._has_gpu:
			self._device = torch.device('cuda')
			self._num_devices = torch.cuda.device_count()
			self._model.to(self._device)

		else:
			self._device = torch.device('cpu')
			self._num_devices = torch.get_num_threads()

		self._model.eval()

	def __repr__(self):
		device = ('GPU' if self._has_gpu else 'CPU')
		return f'<{self.__class__.__name__}_{device}x{self._num_devices}>'

	def __str__(self):
		return repr(self)

	def vectorize(self, text, device=None, get_num_tokens=False):
		"""
		:type text: str or list[str] or Series
		:type device: str or NoneType
		:rtype: numpy.array
		"""
		return self._model.vectorize(text=text, device=device or self._device, get_num_tokens=get_num_tokens)



from itertools import zip_longest
import re
from .Word import Word
from .get_sentence_similarity import get_sentence_similarity

class Sentence:
	def __init__(self, string, similarity_function=None, word_similarity_function=None):
		self._string = str(string)
		if similarity_function is None:
			similarity_function = lambda words1, words2: get_sentence_similarity(
				words1=words1, words2=words2, first_char_weight=0.5, first_word_weight=1.0, case_sensitivity=0.1,
				method='sentence_jaro_winkler'
			)
		self._similarity_function = similarity_function
		self._word_similarity_function = word_similarity_function
		self._words = None


	@property
	def words(self):
		if self._words is None:
			words = [
				Word(string=s, similarity_function=self._word_similarity_function)
				for s in re.findall(r'[^\s!,.?":;]+', self._string)
			]
			self._words = [word for word in words if word.length > 0]

		return self._words


	def __sub__(self, other):
		"""
		:type other: Sentence
		:rtype: float
		"""
		return (1-self.get_similarity(other))*max(self.length, other.length)

	def get_similarity(self, other):
		return self.get_similarity(other=other)

	def __contains__(self, item):
		"""
		:type item: Word or str
		:rtype: bool
		"""
		return str(item) in self._words

	def get_initials(self):
		return [word.initial for word in self._words]

	def get_initial_intersection_score(self, other):
		"""
		:type other: Sentence
		:rtype: int
		"""
		return len(set(self.get_initials()).intersection(set(other.get_initials())))


	def get_word(self, i=1):
		try:
			return self.words[i-1].string
		except:
			return None

	@property
	def length(self):
		return len(self.words)

	def lower(self, inplace=False):
		if inplace:
			self._words = [w.lower(inplace=False) for w in self.words]
			self._string = self._string.lower()
		else:
			return self.__class__(
				string=self._string.lower(), similarity_function=self._similarity_function,
				word_similarity_function=self._word_similarity_function
			)

	def remove_words(self, words, inplace=False):
		"""
		:type words: list of Word or list of str
		"""
		new_words = [w for w in self.words if w not in words]
		new_strings = [w.string for w in new_words]
		if inplace:
			self._words = new_words
			self._string = ' '.join(new_strings)
		else:
			return self.__class__(string=' '.join(new_strings), similarity_function=self._similarity_function)


	def replace_words(self, replacement_dict, inplace=False):
		"""
		:param replacement_dict: a dictionary with words to be replaced as keys and replacements as values
		:type replacement_dict: dict
		"""
		new_strings = [replacement_dict[w.string] if w.string in replacement_dict else w.string for w in self.words]
		if inplace:
			self._words = [Word(string) for string in new_strings]
			self._string = ' '.join(new_strings)
		else:
			return self.__class__(string=' '.join(new_strings), similarity_function=self._similarity_function)

	def __str__(self):
		return self._string

	def __repr__(self):
		words = self.words
		return ' '.join([str(word) for  word in words])


from .GoogleDocumentObject import GoogleDocumentObject


class GoogleToken(GoogleDocumentObject):
	def __init__(self, dictionary, document):
		text = dictionary.pop('text')
		content = text.pop('content')
		begin = text.pop('begin_offset')
		end = begin + len(content)
		super().__init__(dictionary=dictionary, document=document, begin=begin, end=end)

		self._text = content
		self._grammer = self._dictionary.pop('part_of_speech')
		self._part_of_speech = self._grammer.pop('tag', None)
		self._number = self._grammer.pop('number', None)
		self._person = self._grammer.pop('person', None)
		self._index = None

	@property
	def id(self):
		return self.document.id, 'token', self._index

	@property
	def index(self):
		return self._index

	@property
	def begin(self):
		return self._begin

	@property
	def end(self):
		return self._end

	@property
	def text(self):
		return self._text

	@property
	def part_of_speech(self):
		return self._part_of_speech

	@property
	def number(self):
		return self._number

	@property
	def person(self):
		return self._person

	def __repr__(self):
		return str(self._text)

	def __str__(self):
		return f'{self.text}'

	def graph_str(self):
		if self.part_of_speech:
			part_of_speech = f"\n{self.part_of_speech.replace('_', ' ')}"
		else:
			part_of_speech = ''
		return f"[{self.index}] '{self}'" + part_of_speech


	def __eq__(self, other):
		return self._text == other._text and self._begin == other._begin and self._end == other._end



def get_char_by_char_equality(s1, s2):
	if s1 is None or s2 is None: return 0
	s1, s2 = str(s1), str(s2)
	weights = [1 / (2 ** i) for i in range(1, max(len(s1), len(s2)))]
	scores = [(c1==c2)*w for c1, c2, w in zip(s1, s2, weights)]
	total_score = sum(scores)
	return total_score





import torch
import torch.nn as nn
from torch import Tensor
from pandas import Series
from transformers import BertTokenizer
from transformers import BertModel


class BertVectorizerModel(nn.Module):
	def __init__(self, num_tokens=50):
		super().__init__()
		self._bert_layer = BertModel.from_pretrained('bert-base-uncased')
		self._tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
		self._num_tokens = num_tokens

		self._classification_layer = nn.Linear(768, 1)
		self._activation_layer = nn.Sigmoid()

	def forward(self, sequence, attention_mask):
		"""

		:param sequence: Tensor of shape [B, T] containing token ids of sequences
		:param attention_mask: Tensor of shape [B, T] containing attention masks to be used to avoid contribution of PAD tokens
		:return:
		"""

		#  feed the input to BERT to obtain contextualized representations
		contextualized_representations, _ = self._bert_layer(sequence, attention_mask=attention_mask)

		#  obtain the representation of [CLS] head
		classification_representation = contextualized_representations[:, 0]

		return classification_representation

	def clean_text(self, string, get_num_tokens=False):
		"""
		:type string: str
		:rtype: Tensor
		"""
		tokens = ['[CLS]'] + self._tokenizer.tokenize(string) + ['[SEP]']
		num_tokens = len(tokens)

		if len(tokens) > self._num_tokens:
			tokens = tokens[: self._num_tokens - 1] + ['[SEP]']
		else:
			tokens += ['[PAD]'] * (self._num_tokens - len(tokens))

		token_ids = self._tokenizer.convert_tokens_to_ids(tokens)
		# token_ids_tensor = torch.tensor(token_ids)

		result = token_ids
		if get_num_tokens:
			return result, num_tokens
		else:
			return result

	def vectorize(self, text, device=None, get_num_tokens=False):
		"""
		:type text: str or list[str] or Series
		:type device: str or NoneType
		:type get_num_tokens: bool
		:rtype: numpy.array
		"""
		if len(text) == 0:
			if get_num_tokens:
				return None, None
			else:
				return None

		if isinstance(text, (list, Series)):
			_cleaned_text = [self.clean_text(sentence, get_num_tokens=get_num_tokens) for sentence in text]
		elif isinstance(text, str):
			_cleaned_text = [self.clean_text(text, get_num_tokens=get_num_tokens)]
		else:
			raise TypeError(f'text cannot be of type {type(text)}')

		if get_num_tokens:
			cleaned_text = [x[0] for x in _cleaned_text]
			num_tokens = [x[1] for x in _cleaned_text]
		else:
			cleaned_text = _cleaned_text
			num_tokens = None

		sequence = torch.tensor(cleaned_text)
		if device:
			sequence = sequence.to(device)

		attention_mask = torch.stack(
			[(sentence != 0).long() for sentence in sequence]
		)

		if device:
			attention_mask = attention_mask.to(device)

		result = self.forward(sequence=sequence, attention_mask=attention_mask).cpu().data.numpy()
		if get_num_tokens:
			return result, num_tokens
		else:
			return result




















import re
from .get_string_similarity import get_string_similarity


class Word:
	def __init__(self, string, similarity_function=None):
		if string is None: string = ''
		if type(string) is Word:
			string = string._string
		else:
			string = str(string)
		self._string = re.sub(r'\W+', '', string)

		if similarity_function is None:
			similarity_function = lambda s1, s2: get_string_similarity(
				s1=s1, s2=s2, case_sensitivity=0.1, first_char_weight=0.5, method='jaro_winkler'
			)
		self._similarity_function = similarity_function

	@property
	def string(self):
		return self._string

	@property
	def initial(self):
		return self.string[0].upper()

	@property
	def length(self):
		return len(self._string)

	def __sub__(self, other):
		"""
		:type other: Word
		:rtype: float
		"""

		try:
			result = self._similarity_function(self.string, str(other))
		except:
			print(f'{self}, {other}')
			raise
		return result

	def get_similarity(self, other):
		if other is None: return 0
		return self._similarity_function(self.string, str(other))

	def __eq__(self, other):
		return self.string == str(other)

	def lower(self, inplace=False):
		if inplace:
			self._string = self.string.lower()
		else:
			return self.__class__(string=self.string.lower(), similarity_function=self._similarity_function)

	def __repr__(self):
		return self._string

	def __str__(self):
		return self._string




from .DocumentObject import DocumentObject
from spacy import tokens


class TokenSpan(DocumentObject):

	@property
	def character_range(self):
		return self.tokens[0].character_range[0], self.tokens[-1].character_range[1]

	def __str__(self):
		character_range = self.character_range
		return self.document.text[character_range[0]: character_range[1]]
		# return ' '.join([str(x) for x in self.tokens])

	def __repr__(self):
		return str(self)

	@property
	def obj(self):
		"""
		:rtype: tokens.Span
		"""
		return self._obj

	@property
	def start(self):
		"""
		:rtype: int
		"""
		try:
			return self._start
		except AttributeError:
			self._start = self.obj.start
		return self._start

	@property
	def end(self):
		"""
		:rtype: int
		"""
		try:
			return self._end
		except AttributeError:
			self._end = self.obj.end
		return self._end

	@property
	def end_except_punctuation(self):
		"""
		:rtype: int
		"""
		try:
			return self._end_except_punctuation
		except AttributeError:
			self._end_except_punctuation = max([x.index for x in self.tokens if not x.is_punctuation])
		return self._end_except_punctuation

	@property
	def id(self):
		return (self.document.id, 'span', self.start, self.end)

	@property
	def text(self):
		"""
		:rtype: str
		"""
		return self.obj.text


	@property
	def tokens(self):
		"""
		:rtype: list[.Document.Token]
		"""
		return self.document.tokens[self.start:self.end]

	@property
	def last_token(self):
		"""
		:rtype: .Document.Token
		"""
		return self.tokens[-1]

	@property
	def is_sentence(self):
		try:
			return self._is_sentence

		except AttributeError:
			self._is_sentence = False
			for sentence in self.document.sentences:
				if sentence.start == self.start and sentence.end_except_punctuation == self.end_except_punctuation:
					self._is_sentence = True
					break

		return self._is_sentence

	def has(self, part_of_speech):
		return any([token.part_of_speech == part_of_speech for token in self.tokens])



from abstract import Graph
from abstract.graph_style import NodeStyle

from .GoogleToken import GoogleToken
from .GoogleEntity import GoogleEntity
from .GoogleSentence import GoogleSentence
from .GoogleSentiment import GoogleSentiment


class GoogleDocument:
	def __init__(self, text=None, cloud=None, id=0, _analysis=None):
		"""
		:param str or NoneType text: a text to be converted to a GoogleDocument
		:param .GoogleCloud.GoogleCloud cloud: GoogleCloud API
		:param int or str id:
		:param dict or NoneType _analysis: result of GoogleCloud.analyze(text)
		"""

		if isinstance(text, list):
			text = ' '.join([str(x) for x in text])
		self._id = id

		if text is None and _analysis is None:
			raise ValueError('Either text or _analysis should be given!')
		elif text is None:
			self._dictionary = _analysis.copy()
			self._cloud = cloud
		elif _analysis is None:
			self._cloud = cloud
			self._dictionary = self._cloud.analyze(text=text)

		self._text = self._dictionary.pop('text')
		self._sentences = None
		self._tokens = None
		self._entities = None
		sentiment = self._dictionary.pop('document_sentiment')
		self._sentiment = GoogleSentiment(score=sentiment.pop('score'), magnitude=sentiment.pop('magnitude'))
		self._language = None
		self._graph = None

	def __str__(self):
		return '\n'.join([str(sentence) for sentence in self.sentences])

	def __repr__(self):
		return str(self)

	@property
	def id(self):
		return self._id

	@property
	def cloud(self):
		"""
		:rtype: .GoogleCloud.GoogleCloud
		"""
		return self._cloud

	def _get_from_dictionary(self, name):
		if name in self._dictionary:
			return self._dictionary.pop(name)
		elif name in ['text', 'sentiment', 'language']:
			return False
		else:
			return []

	def tokenize(self):
		self._text = self._get_from_dictionary('text')

		self._sentences = [GoogleSentence(x, document=self) for x in self._get_from_dictionary('sentences')]
		self._sentences.sort()
		for index, sentence in enumerate(self._sentences):
			sentence._index = index

		self._tokens = [GoogleToken(x, document=self) for x in self._get_from_dictionary('tokens')]
		self._tokens.sort()
		for index, token in enumerate(self._tokens):
			token._index = index

		self._entities = [GoogleEntity(x, document=self) for x in self._get_from_dictionary('entities')]
		for index, entity in enumerate(self._entities):
			entity._index = index

		self._sentiment = self._get_from_dictionary('sentiment')

		self._language = self._get_from_dictionary('language')

	@property
	def text(self):
		if self._text is None:
			self.tokenize()
		return self._text

	@property
	def sentences(self):
		"""
		:rtype: list[GoogleSentence]
		"""
		if self._sentences is None:
			self.tokenize()
		return self._sentences

	@property
	def tokens(self):
		"""
		:rtype: list[GoogleToken]
		"""
		if self._tokens is None:
			self.tokenize()
		return self._tokens

	@property
	def entities(self):
		"""
		:rtype: list[GoogleEntity]
		"""
		if self._entities is None:
			self.tokenize()
		return self._entities

	@property
	def sentiment(self):
		"""
		:rtype: GoogleSentiment
		"""
		if self._sentiment is None:
			self.tokenize()
		return self._sentiment

	@property
	def language(self):
		if self._language is None:
			self.tokenize()
		return self._language

	def graph_str(self):
		return '\n'.join([sentence.graph_str() for sentence in self.sentences])

	@property
	def graph(self):
		"""
		:rtype: Graph
		"""
		sentence_style = NodeStyle(text_size=7, shape='rect', style='"rounded, filled"')
		entity_style = NodeStyle(fill_colour='gold3', text_colour='black')
		mention_style = NodeStyle(fill_colour='gold', text_colour='black')

		if self._graph is None:
			self._graph = Graph()
			for token in self.tokens:
				self._graph.add_node(name=str(token.id), label=token.graph_str())

			if len(self._sentences) > 1:
				self._graph.add_node(
					name=str(self.id),
					label=f'{self.graph_str()}\n({len(self.sentences)} sentences)',
					style=sentence_style
				)

				for index, sentence in enumerate(self.sentences):
					self._graph.add_node(
						name=str(sentence.id),
						label=f'{sentence.graph_str()}\n(sentence {index+1})',
						style=sentence_style
					)
					self._graph.connect(start=str(self.id), end=str(sentence.id))  # document --> sentence
					for token in sentence.tokens:
						self._graph.connect(start=str(sentence.id), end=str(token.id))  # sentence --> token

			else:
				self._graph.add_node(
					name=str(self.id),
					label=f'{self.graph_str()}\n({len(self.sentences)} sentence)',
					style=sentence_style
				)
				for sentence in self.sentences:
					for token in sentence.tokens:
						self._graph.connect(start=str(self.id), end=str(token.id))  # document --> token

			for entity in self.entities:
				self._graph.add_node(
					name=str(entity.id),
					label=entity.graph_str(),
					style=entity_style
				)

				for index, mention in enumerate(entity.mentions):
					self._graph.add_node(
						name=str(mention.id),
						label=f'mention {index+1}',
						style=mention_style
					)
					self._graph.connect(start=str(mention.id), end=str(entity.id))  # entity <-- mention
					for token in mention.tokens:
						self._graph.connect(start=str(token.id), end=str(mention.id))  # mention <-- token

		return self._graph




class GoogleSentiment:
	def __init__(self, score, magnitude):
		self._score = score
		self._magnitude = magnitude

	@property
	def score(self):
		return self._score

	@property
	def magnitude(self):
		return self._magnitude



# -*- coding: utf-8 -*-
import re
alphabets= "([A-Za-z])"
prefixes = "(Mr|St|Mrs|Ms|Dr)[.]"
suffixes = "(Inc|Ltd|Jr|Sr|Co)"
starters = "(Mr|Mrs|Ms|Dr|He\s|She\s|It\s|They\s|Their\s|Our\s|We\s|But\s|However\s|That\s|This\s|Wherever)"
acronyms = "([A-Z][.][A-Z][.](?:[A-Z][.])?)"
websites = "[.](com|net|org|io|gov)"

def split_into_sentences(text):
    text = " " + text + "  "
    text = text.replace("\n"," ")
    text = re.sub(prefixes,"\\1<prd>",text)
    text = re.sub(websites,"<prd>\\1",text)
    if "Ph.D" in text: text = text.replace("Ph.D.","Ph<prd>D<prd>")
    text = re.sub("\s" + alphabets + "[.] "," \\1<prd> ",text)
    text = re.sub(acronyms+" "+starters,"\\1<stop> \\2",text)
    text = re.sub(alphabets + "[.]" + alphabets + "[.]" + alphabets + "[.]","\\1<prd>\\2<prd>\\3<prd>",text)
    text = re.sub(alphabets + "[.]" + alphabets + "[.]","\\1<prd>\\2<prd>",text)
    text = re.sub(" "+suffixes+"[.] "+starters," \\1<stop> \\2",text)
    text = re.sub(" "+suffixes+"[.]"," \\1<prd>",text)
    text = re.sub(" " + alphabets + "[.]"," \\1<prd>",text)
    if "”" in text: text = text.replace(".”","”.")
    if "\"" in text: text = text.replace(".\"","\".")
    if "!" in text: text = text.replace("!\"","\"!")
    if "?" in text: text = text.replace("?\"","\"?")
    text = text.replace(".",".<stop>")
    text = text.replace("?","?<stop>")
    text = text.replace("!","!<stop>")
    text = text.replace("<prd>",".")
    sentences = text.split("<stop>")
    sentences = sentences[:-1]
    sentences = [s.strip() for s in sentences]
    return sentences


from .levenshtein import get_levenshtein_similarity
from .jaro_winkler import get_jaro_winkler_similarity




def get_string_similarity(s1, s2, case_sensitivity=1.0, first_char_weight=0.0, method='jaro_winkler'):


	s1 = s1 or ''
	s2 = s2 or ''
	s1 = str(s1)
	s2 = str(s2)

	if method=='jaro_winkler':
		_get_similarity = get_jaro_winkler_similarity
	elif method=='levenshtein':
		_get_similarity = get_levenshtein_similarity
	else:
		raise ValueError(f'{method} is unknown!')

	if first_char_weight<0:
		raise ValueError('initial_weight cannot be negative!')

	if case_sensitivity==0:
		string_similarity = _get_similarity(s1, s2, case_sensitive=False)
		if first_char_weight>0 and len(s1)>0 and len(s2)>0:
			initials_equal = int(s1[0].lower() == s2[0].lower())
		else:
			initials_equal = 0
		similarity = (string_similarity + initials_equal * first_char_weight) / (1 + first_char_weight)

	elif case_sensitivity==1:
		string_similarity = _get_similarity(s1, s2, case_sensitive=True)
		if first_char_weight>0 and len(s1)>0 and len(s2)>0:
			initials_equal = int(s1[0] == s2[0])
		else:
			initials_equal = 0
		similarity = (string_similarity + initials_equal * first_char_weight) / (1 + first_char_weight)

	elif case_sensitivity>0 and case_sensitivity<1:
		sensitive = _get_similarity(s1, s2, case_sensitive=True)
		insensitive = _get_similarity(s1, s2, case_sensitive=False)
		string_similarity = case_sensitivity*sensitive + (1-case_sensitivity)*insensitive
		if first_char_weight>0 and len(s1)>0 and len(s2)>0:
			initials_equal = case_sensitivity*int(s1[0] == s2[0]) + (1-case_sensitivity)*int(s1[0].lower() == s2[0].lower())
		else:
			initials_equal = 0
		similarity = (string_similarity + initials_equal * first_char_weight) / (1 + first_char_weight)

	else:
		raise ValueError('case_sensitivity should be between 0 and 1.')

	return similarity



from editdistance import eval as get_edit_distance
from jellyfish import jaro_winkler as get_similarity

def convert_distance_function_to_score_function(function=get_edit_distance):
	def score_func(s1, s2):
		distance = function(s1, s2)
		score = max(0,1-float(distance)/max(len(s1), len(s2)))
		return score
	return score_func

def create_weighted_similarity_function(function=get_similarity, first_letter_weight=1):
	def new_score_func(s1, s2):
		original_result = function(s1, s2)
		first_letter_result = float(s1[0]==s2[0])
		return (original_result+first_letter_result*first_letter_weight)/(1.0+first_letter_weight)
	return new_score_func